/************************************************************************************
iPhone Communication RT-Component
Copyright (c) 2010, Nagoya Institute of Technology
All rights reserved.

Contact us if you use this software for sell.
If you use this software not for sell, you can use this software under BSD lisence.
See the files LICENSE.TXT and LICENSE-BSD.TXT for more details.                     
************************************************************************************/
import javax.vecmath.*;
import java.awt.geom.*;

public class VirtualTargetFollowing {



	public double kv=0.0005;
	public double kw=0.6;
	public double ktheta=0.5;
	public double kdist=0.0;
	public int forwardPointNum=1;

	public Point2d getVelocity(Point3d robotPosition, Point2d[] trajectory, Point2d preVelocity,Point2d[] lrfDataRobot){

		Point2d velocity =new Point2d(0,0);

		Point2d posXY = new Point2d(robotPosition.x, robotPosition.y);
		double minDistance =10000000;
		int minID=0;
		int targetID=0;

		//軌道上でロボットに一番近い点を求める
		for(int i=0;i<trajectory.length;i++){
			double distance = posXY.distance(trajectory[i]);
			if(distance<minDistance){
				minDistance = distance;
				minID=i;
			}
		}
//		System.out.println("VirtualTarget:"+minDistance+", "+minID);

		//一番近い点の5個先に仮想ターゲットをセット
		//終点間際にはつねに終点に仮想ターゲットをセット
		if(minID>trajectory.length-forwardPointNum){
			targetID=trajectory.length-1;
		}else{
			targetID=minID+forwardPointNum-1;
		}
//		System.out.println("VirtualTarget:"+targetID);

		Point2d virtualTarget = trajectory[targetID];

		double targetDistance=0;
		double targetAngle=0;
		targetDistance= posXY.distance(virtualTarget);
		targetAngle = Math.atan2(virtualTarget.y-posXY.y, virtualTarget.x-posXY.x)+robotPosition.z;
		System.out.println("target:"+targetDistance+", "+Math.toDegrees(targetAngle));


		if(targetAngle>=Math.PI)	targetAngle-=2*Math.PI;
		if(targetAngle<=-Math.PI)	targetAngle+=2*Math.PI;
//		System.out.println("new targetAngle(deg):"+Math.toDegrees(targetAngle));


		double v=kv*targetDistance;
		double w=-kw*preVelocity.y+ktheta*targetAngle+kdist*targetDistance;
		if(v>0.5)	v=0.5;
		if(w>0.5)	w=0.5;
		if(Math.abs(targetAngle)>Math.toRadians(30.0))	v=0.0;
		velocity=new Point2d(v,w);



		//ターゲット付近に到着したら停止
		int finishID=(int)((double)trajectory.length*0.8);
		System.out.println("ID:"+minID+"/"+trajectory.length+"  ("+finishID+")");
		boolean distFlag=false;
		if(minID>=finishID){
			distFlag=true;
//			System.out.println("up!!!");
		}
		if(targetDistance<50 && distFlag){
//			System.out.println("up2!!!!!!!!!!!!!!!");
			velocity=new Point2d(0,0);
		}



		//進行方向に何かあれば並進速度をゼロにする
		for(int i=0;i<lrfDataRobot.length;i++){
//			System.out.println("lrfDataRobot["+i+"]:"+lrfDataRobot[i]);
			if(lrfDataRobot[i].x >0 &&lrfDataRobot[i].x <400 && Math.abs(lrfDataRobot[i].y)<340){
				velocity = new Point2d(0, velocity.y);
//				System.out.println("  -> obstacle");
			}
		}

//		System.out.println("v,w:"+velocity);



		return velocity;

	}
	public Point2d getVelocityForiPhone(Point3d robotPosition, Point2d targetPoint, Point2d preTargetPoint, Point2d preVelocity){
		Point2d velocity;
		double v=0;
		double w=0;
		double kw=0.001;
		double ktheta=3.0;
		double kdist=0.0001;
		double targetDistance=0;
		double targetAngle=0;
		
		Point2d posXY = new Point2d(robotPosition.x, robotPosition.y);

		targetDistance= posXY.distance(targetPoint);
//		targetAngle = Math.atan2(targetPoint.y-posXY.y, targetPoint.x-posXY.x)+robotPosition.z;
		targetAngle = Math.atan2(targetPoint.y-posXY.y, targetPoint.x-posXY.x)-robotPosition.z;
		if(targetAngle>=Math.PI)	targetAngle-=2*Math.PI;
		if(targetAngle<=-Math.PI)	targetAngle+=2*Math.PI;
		System.out.println("target(dis,ang):"+targetDistance+", "+Math.toDegrees(targetAngle));
		
		Line2D.Double targetLine=new Line2D.Double(targetPoint.x,targetPoint.y,preTargetPoint.x,preTargetPoint.y);
		double distance=targetLine.ptLineDist(posXY.x, posXY.y);
		System.out.println("distance between robot and line:"+distance);
		
		w=preVelocity.y-(kw*preVelocity.y+ktheta*targetAngle+kdist*distance)*0.1;
		System.out.println("w: "+w);
		if(w<-0.5)	w=-0.5;
		if(w>0.5)	w=0.5;
		if(Math.abs(targetAngle)>Math.toRadians(30.0)){
			v=0.0;
		}else{
			v=0.3;
		}
		velocity=new Point2d(v,-w);
		System.out.println("velocity: "+velocity);
			
		return velocity;
	}
	

}
